#lang racket/base

(require "model.rkt"
         "util/rm.rkt"
         (for-syntax racket/base)
         redex
         racket/contract
         tests/eli-tester
         racket/match
         racket/unit
         racket/runtime-path)

(define-signature agda-types-signature
  (infer-type frp->agda save complex generate
              sel))
  
(define-unit agda-types
  (import)
  (export agda-types-signature)
  
  (define (print-SF-type SF port mode)
    (define (print-type type)
      (match type
        [(event type) (format "( E ~a )" type)]
        [(continuous type) (format "( C ~a )" type)]))
    (define (print-types types)
      (local-require racket/list)
      (define out (add-between (map print-type types) " ∷ "))
      (apply string-append (append out (list " ∷ []"))))
    (fprintf port "SF ( ~a ) ( ~a ) ~a"
             (print-types (SF-type-inputs SF))
             (print-types (SF-type-outputs SF))
             (SF-type-kind SF)))
  
  (define (sel x) (SF-type-inputs x))
                                   
  (struct frp-event () #:transparent)
  (struct event frp-event (type) #:transparent)
  (struct continuous frp-event (type) #:transparent)
  (define-struct/contract SF-type ([inputs (non-empty-listof frp-event?)]
                                   [outputs (non-empty-listof frp-event?)]
                                   [kind (or/c 'cau 'dec)])
    #:property prop:custom-write print-SF-type
    #:transparent)
  
  (define-syntax (define-SF-type stx)
    (define (parse-type type)
      (syntax-case type (E C)
        [(E type) #'(event `type)]
        [(C type) #'(continuous `type)]))
    (syntax-case stx (SF ::)
      [(_ name :: SF input output kind)
       (with-syntax ([input-type (parse-type #'input)]
                     [output-type (parse-type #'output)])
         #'(define name (SF-type (list input-type)
                                 (list output-type)
                                 'kind)))]))

  (define integer 'ℕ)
  (define-SF-type dprim1-SF-type :: SF [C ,integer] [C ,integer] dec)
  (define-SF-type dprim2-SF-type :: SF [E ,integer] [C ,integer] dec)
  (define-SF-type dprim3-SF-type :: SF [C ,integer] [E ,integer] dec)
  (define-SF-type dprim4-SF-type :: SF [E ,integer] [E ,integer] dec)
  
  (define-SF-type prim1-SF-type :: SF [E ,integer] [E ,integer] cau)
  (define-SF-type prim2-SF-type :: SF [C ,integer] [E ,integer] cau)
  (define-SF-type prim3-SF-type :: SF [E ,integer] [C ,integer] cau)
  (define-SF-type prim4-SF-type :: SF [C ,integer] [C ,integer] cau)
  
  (define-SF-type switch-function1-type :: SF [E integer] [E integer] dec)
  (define-SF-type switch-function2-type :: SF [E integer] [E integer] dec)
  (define-SF-type switch-function3-type :: SF [E integer] [E integer] dec)
  
  (define (combine-kind-or kind1 kind2)
    (match (list kind1 kind2)
      [(list 'dec 'dec) 'dec]
      [(list-no-order 'dec 'cau) 'cau]
      [(list 'cau 'cau) 'cau]))
  
  (define (combine-kind-and kind1 kind2)
    (match (list kind1 kind2)
      [(list 'dec 'dec) 'dec]
      [(list-no-order 'dec 'cau) 'dec]
      [(list 'cau 'cau) 'cau]))
  
  (define (infer-type frp)
    (redex/match FRP:restricted frp
                 [(dprim dprim-function state)
                  (case (term dprim-function)
                    [(dprim1) dprim1-SF-type]
                    [(dprim2) dprim2-SF-type]
                    [(dprim3) dprim3-SF-type]
                    [(dprim4) dprim4-SF-type]
                    [else (error 'infer-type "invalid dprim function ~a" (term dprim-function))])]
                 [(prim prim-function state)
                  (case (term prim-function)
                    [(prim1) prim1-SF-type]
                    [(prim2) prim2-SF-type]
                    [(prim3) prim3-SF-type]
                    [(prim4) prim4-SF-type]
                    [else (error 'infer-type "invalid prim function ~a" (term prim-function))])]
                 [(SF_1 &&& SF_2)
                  (define SF_1-type (infer-type (term SF_1)))
                  (define SF_2-type (infer-type (term SF_2)))
                  (SF-type (SF-type-inputs SF_1-type)
                           (append (SF-type-outputs SF_1-type)
                                   (SF-type-outputs SF_2-type))
                           (combine-kind-and (SF-type-kind SF_1-type)
                                             (SF-type-kind SF_2-type)))]
                 [(SF_1 >>> SF_2)
                  (define SF_1-type (infer-type (term SF_1)))
                  (define SF_2-type (infer-type (term SF_2)))
                  (SF-type (SF-type-inputs SF_1-type)
                           (SF-type-outputs SF_2-type)
                           (combine-kind-and (SF-type-kind SF_1-type)
                                             (SF-type-kind SF_2-type)))]
                 [(loop SF_1 SF_2)
                  (define SF_1-type (infer-type (term SF_1)))
                  (SF-type (list (car (SF-type-inputs SF_1-type)))
                           (list (car (SF-type-outputs SF_1-type)))
                           (SF-type-kind SF_1-type))]
                 [(switch SF switch-function)
                  (define type (infer-type (term SF)))
                  (define switch-type (infer-type (term switch-function)))
                  (SF-type (SF-type-inputs type)
                           (SF-type-outputs switch-type)
                           (combine-kind-or (SF-type-kind type)
                                            (SF-type-kind switch-type)))]
                 [(dswitch SF switch-function)
                  (define type (infer-type (term SF)))
                  (define switch-type (infer-type (term switch-function)))
                  (SF-type (SF-type-inputs type)
                           (SF-type-outputs switch-type)
                           (combine-kind-or (SF-type-kind type)
                                            (SF-type-kind switch-type)))]
                 [(SF_1 *** SF_2)
                  (define SF_1-type (infer-type (term SF_1)))
                  (define SF_2-type (infer-type (term SF_2)))
                  (SF-type (append (SF-type-inputs SF_1-type)
                                   (SF-type-inputs SF_2-type))
                           (append (SF-type-outputs SF_1-type)
                                   (SF-type-outputs SF_2-type))
                           (combine-kind-or (SF-type-kind SF_1-type)
                                            (SF-type-kind SF_2-type)))]
                 ))

  (define-syntax (pick-one stx)
    (syntax-case stx ()
      [(_ stuff ...)
       (let ()
         (define things (length (syntax->list #'(stuff ...))))
         (with-syntax ([total things]
                       [(number ...) (for/list ([i (in-range things)]) i)])
           #'(case (random total)
               [(number) stuff]
               ...)))]))

  (define (do-generate input output kind depth #:max-depth [maximum 1])
    (define (choose-type)
      (pick-one (event 'integer)
                (continuous 'integer)))
    (define (make->>> input output kind)
      (define new-output (choose-type))
      (let ([term1 (do-generate input new-output kind (add1 depth))]
            [term2 (do-generate new-output output kind (add1 depth))])
        (term (,term1 >>> ,term2))))
    (define (make-loop input output kind)
      (define next-depth (add1 depth))
      (define loop-sfs-bs output)
      (define loop-sfs-ds (choose-type))
      (define loop-sfs-cs (choose-type))
      (define term2-output loop-sfs-cs)
      (define term1 (do-generate input output kind next-depth))
      (define term2 (do-generate input term2-output kind next-depth))
      (define loop1 (do-generate (list output loop-sfs-cs)
                                 (list loop-sfs-bs loop-sfs-ds)
                                 kind next-depth))
      (define loop2 (do-generate loop-sfs-ds output 'dec next-depth))
      (term ((,term1 &&& ,term2) >>> (loop ,loop1 ,loop2))))

    (define (normalize stuff)
      (match stuff
        [(list one) one]
        [else stuff]))

    (define (split stuff)
      (local-require racket/list)
      (values
        (normalize (take stuff (quotient (length stuff) 2)))
        (normalize (drop stuff (quotient (length stuff) 2)))))

    (define (make-*** inputs outputs kind)
      (define-values (input-1 input-2) (split inputs))
      (define-values (output-1 output-2) (split outputs))
      (define left (do-generate input-1 output-1 kind (add1 depth)))
      (define right (do-generate input-2 output-2 kind (add1 depth)))
      (term (,left *** ,right)))

    (define (need-atomic?)
      (match (list input output depth)
        [(list (or (event _)
                   (continuous _))
               (or (event _)
                   (continuous _))
               (? (lambda (x) (> x maximum))))
         #t]
        [else #f]))

    (define (atomic)
      (match (list input output kind)
        [(list (event _)
               (event _)
               'cau)
         (pick-one
           (term (prim prim1 4)))]
        [(list (event _)
               (event _)
               'dec)
         (pick-one
           (term (dprim dprim4 4)))]
        [(list (event _)
               (continuous _)
               'cau)
         (pick-one
           (term (prim prim3 4)))]
        [(list (event _)
               (continuous _)
               'dec)
         (pick-one
           (term (dprim dprim2 4)))]
        [(list (continuous _)
               (continuous _)
               'cau)
         (pick-one
           (term (prim prim3 2)))]
        [(list (continuous _)
               (continuous _)
               'dec)
         (pick-one
           (term (dprim dprim1 3)))]
        [(list (continuous _)
               (event _)
               'dec)
         (pick-one
           (term (dprim dprim3 9)))]
        [(list (continuous _)
               (event _)
               'cau)
         (pick-one
           (term (prim prim2 10)))]
        ))

    (define (maybe-atomic)
      (if (match (list input output)
            [(list (or (event _)
                       (continuous _))
                   (or (event _)
                       (continuous _))) #t]
            [else #f])
        (atomic)
        (do-generate input output kind depth)))

    (define next-depth (add1 depth))

    (if (need-atomic?)
      (atomic)
      (match (list input output kind)
        [(list (or (event _)
                   (continuous _))
               (or (event _)
                   (continuous _))
               kind)
         (pick-one
           (atomic)
           (make->>> input output kind)
           (make-loop input output kind))]
        [(list (list inputs ..2)
               (list outputs ..2)
               kind)
         (pick-one
           (make-*** inputs outputs kind))]
        [else (error 'generate "input ~a output ~a kind ~a depth ~a" input output kind depth)])))

  (define (generate)
    (do-generate (pick-one (event 'integer)
                           (continuous 'integer))
                 (pick-one (event 'integer)
                           (continuous 'integer))
                 (pick-one 'cau 'dec)
                 0))
    
  (test (infer-type (term (dprim dprim1 2))) => dprim1-SF-type
        (infer-type (term ((dprim dprim1 2) *** (dprim dprim1 5))))
        =>
        (SF-type
         (list (continuous 'ℕ) (continuous 'ℕ))
         (list (continuous 'ℕ) (continuous 'ℕ))
         'dec))
  
  (define (complex frp)
    (redex/match FRP:restricted frp
                 [(dprim dprim-function state) #f]
                 [(prim prim-function state) #f]
                 [SF #t]))
  
  (define (new-variable) (gensym 'frp))
  
  (define (frp->agda frp)
    (define (implementation frp)
      (redex/match FRP:restricted frp
                   [(dprim dprim-function state)
                    (format "(~a ~a)" (term dprim-function) (term state))]
                   [(SF_1 op SF_2)
                    (format "(~a ~a ~a)"
                            (implementation (term SF_1))
                            (term op)
                            (implementation (term SF_2)))]
                   [(prim prim-function state)
                    (format "(~a ~a)" (term prim-function) (term state))]
                   [(switch SF switch-function)
                    (format "(switch (~a &&& ~a) (switchwrap ~a))"
                            "maybe_c"
                            (implementation (term SF)) 
                            (implementation (term switch-function)))]
                   [(dswitch SF switch-function)
                    (format "(dswitch (~a &&& ~a) (switchwrap ~a))"
                            "maybe_c"
                            (implementation (term SF))
                            (implementation (term switch-function)))]
                   [(loop SF_1 SF_2)
                    (format "(loop ~a ~a)" (implementation (term SF_1))
                            (implementation (term SF_2)))]))
    (define variable (new-variable))
    (format "~a : ~a\n~a = ~a\n"
            variable (infer-type frp)
            variable (implementation frp)))
  
  #;
  (test (and (regexp-match
              (string-append
               "frp[0-9]+"
               (regexp-quote " : SF ( ( C ℕ ) ∷ [] ) ( ( C ℕ ) ∷ [] ) dec\n")
               "frp[0-9]+"
               (regexp-quote " = (dprim1 2)\n"))
              (frp->agda (term (dprim dprim1 2))))
             #t)
        => #t)

       
  (define (save filename term)
    (with-output-to-file filename #:exists 'replace
      (lambda ()
        (printf "
{-# OPTIONS --type-in-type
 #-}

open import NeilPrelude
open import Core
open import Primitives
open import List
open import Evaluator

module Test1 where

prim1 : ℕ → SF [ E ℕ ] [ E ℕ ] cau
prim1 z = pureE (λ x -> z)

prim2 : ℕ → SF [ C ℕ ] [ E ℕ ] cau
prim2 z = prim (λ t s blah -> unit & ⟨ just z ⟩) unit

prim3 : ℕ → SF [ E ℕ ] [ C ℕ ] cau
prim3 z = prim (λ t s blah -> unit & hvwrap z) unit

prim4 : ℕ → SF [ C ℕ ] [ C ℕ ] cau
prim4 z = prim (λ t s blah -> unit & blah) unit

dprim1 : ℕ → SF [ C ℕ ] [ C ℕ ] dec
dprim1 n = constant n

dprim2 : ℕ → SF [ E ℕ ] [ C ℕ ] dec
dprim2 n = constant n
  
dprim3 : ℕ → SF [ C ℕ ] [ E ℕ ] dec
dprim3 n = dprim (λ t s -> (λ blah ->  unit) & ⟨ just n ⟩) unit 

dprim4 : ℕ → SF [ E ℕ ] [ E ℕ ] dec
dprim4 n = dprim (λ t s -> (λ blah -> unit) & ⟨ just n ⟩) unit

maybe_c : SF [ C ℕ ] [ E Unit ] cau
maybe_c = prim (λ _ _ _ -> unit & ⟨ just unit ⟩) unit

maybe_e : SF [ E ℕ ] [ E Unit ] cau
maybe_e = prim (λ _ _ _ -> unit & ⟨ just unit ⟩) unit

switchwrap : {e : Set} -> {as bs : SVDesc} -> {d : Dec} -> SF as bs d -> (e -> SF as bs d)
switchwrap f = (λ _ -> f)
       
~a
\n" (frp->agda term)))))
    
)  

(define-values/invoke-unit/infer (export (prefix agda: agda-types-signature)) agda-types)

(define (md5sum file)
  (local-require file/md5)
  (with-input-from-file file (lambda ()
                               (md5 (current-input-port)))))

(define (cat filename)
  (with-input-from-file filename (lambda ()
                                   (printf "~a\n" (read-string 100000)))))

(define (delete-file/safe file)
  (local-require racket/file)
  (ignore-errors (when (file-exists? file)
                   (delete-file file))))

(define-runtime-path tests "tests")
(define (generate depth)
  (define (save-term file term)
    (with-output-to-file file #:exists 'replace
      (lambda () (printf "~a\n" term))))
  (define (get)
    (define generated (generate-term FRP:restricted SF depth))
    (if (agda:complex generated)
      generated
      (get)))
  (define name (format "Test-~a" (gensym)))
  (printf "Saving ~a\n" name)
  (define generated (get))
  (unless (directory-exists? tests)
    (make-directory tests))
  (agda:save (build-path tests (format "~a.agda" name)) generated)
  (save-term (build-path tests (format "~a.rktl" name)) generated))


(define (write+run)
  (local-require racket/system)
  (define generated (agda:generate))
  (define agda-ip "192.168.40.35")
  (define name (format "Test-~a.agda" (gensym)))
  (delete-file/safe "agda/Test1.agdai")
  (agda:save "agda/Test1.agda" generated)
  (agda:save name generated)
  (printf "Running ~a ~a " name generated)
  (if (agda:complex generated)
      (let ()
        (local-require racket/port)
        (define command (format "ssh ~a '~a'" agda-ip "(cd redex/code; ./run)"))
        (define-values (stdout stdin pid stderr waiter)
                       (apply values (process command)))
        (waiter 'wait)
        (define code (waiter 'exit-code))
        (printf ": ~a\n" (if (= 0 code) 'ok 'failed))
        (printf "Stdout: ~a\n" (port->string stdout))

        (if (= 0 code)
            generated
            #f))
      #f))

(define-syntax-rule (ignore-errors body ...)
  (with-handlers ([exn? (lambda (e) (printf "~a\n" (exn-message e)))])
    body ...))

(let ()
  (define f1 (frp:restricted->frp (term (loop (loop ((prim prim1 5) *** (dprim dprim1 5))
                                                    (loop (dprim dprim1 2) (dprim dprim1 2)))
                                              (dprim dprim1 5)))))
  (define f2 (frp:restricted->frp (term ((prim prim1 5) *** (dprim dprim1 5)))))
  (define f3 (frp:restricted->frp (term (prim prim1 5))))
  (define f4 (frp:restricted->frp (term (dprim dprim1 5))))
  (define signals '((1) (2) (3) (4)))
  
  (test (evaluate-steps:paper f3 signals)
        =>
        '((5) (5) (5) (5))
        (evaluate-steps:paper f4 signals)
        =>
        '((5) (5) (5) (5))
        
        (evaluate-steps:paper f2 '((1 2)))
        =>
        '((5 5))
        
        (evaluate-steps:paper f1 signals)
        =>
        '((5) (5) (5) (5))))

