#lang racket

(require redex/reduction-semantics racket/trace)
(provide CCA cc-normal reduced weak-head-normal ccnf?)

(define-language CCA
  ((x y z) variable-not-otherwise-mentioned)
  (t unit num)
  ((α β θ) t
           (pair α β)
           (-> α β)
           (arr α β))
  ((e i j f g h)
   x
   (pair e e)
   (fst e)
   (snd e)
   (λ (x) e) ; drop type annotation
   (e e)
   ()
   
   constant)
  (Γ ((x_0 α_0) ...))
  
  (constant arr >>> first
            loop init
            
            real
            True False
            + * /
            
            assoc assoc-inv
            swap times
            compose id
            juggle
            shuffle shuffle-inv
            second
            
            ; What does `loopB' mean on the LHS of the relation clauses?
            ; Currently treating it as an irreducible blackbox (like the
            ; arrow primitives).
            loopB)
  (ccnf (arr e)
        ((loopB e) (arr e))))

(define ↦
  (reduction-relation
   CCA #:domain e
   (--> (loop f)
        ((loopB ()) ((>>> (arr assoc-inv)) ((>>> (first f)) (arr assoc))))
        "loop")
   (--> (init i)
        ((loopB i) (arr ((compose swap) ((compose juggle) swap))))
        "init")
   (--> ((>>> (arr f)) (arr g))
        (arr ((compose g) f))
        "composition")
   (--> (first (arr f))
        (arr ((times f) id))
        "extension")
   (--> ((>>> h) ((loopB i) f))
        ((loopB i) ((>>> (first h)) f))
        "left tightening")
   (--> ((>>> ((loopB i) f)) (arr g))
        ((loopB i) ((>>> f) (first (arr g))))
        "right tightening")
   (--> ((loopB i) ((loopB j) f))
        ((loopB (pair i j)) ((>>> (arr shuffle)) ((>>> f) (arr shuffle-inv))))
        "vanishing")
   (--> (first ((loopB i) f))
        ((loopB i) ((>>> (arr juggle)) ((>>> (first f)) (arr juggle))))
        "superposing")))

(define ⇓
  (reduction-relation
   CCA #:domain e
   (--> ccnf 
        ccnf
        "norm")
   (--> ((>>> e_1) e_2)
        e_’
        (where e_1’ ,(cc-normal (term e_1)))
        (where e_2’ ,(cc-normal (term e_2)))
        (where (e) ,(reduced (term ((>>> e_1’) e_2’))))
        (where e_’ ,(cc-normal (term e)))
        "seq")
   (--> (first f)
        e_’
        (where f_’ ,(cc-normal (term f)))
        (where (e) ,(reduced (term (first f_’))))
        (where e_’ ,(cc-normal (term e)))
        "first")
   (--> (init i)
        e_’
        (where (e) ,(reduced (term (init i))))
        (where e_’ ,(cc-normal (term e)))
        "init")
   (--> (loop f)
        e_’
        (where (e) ,(reduced (term (loop f))))
        (where e_’ ,(cc-normal (term e)))
        "loop")
   (--> ((loopB i) f)
        e_’
        (where f_’ ,(cc-normal (term f)))
        (where (e) ,(reduced (term ((loopB i) f_’))))
        (where e_’ ,(cc-normal (term e)))
        "loopB")
   (--> ((loopB i) f)
        ((loopB i) (arr g))
        (where (arr g) ,(cc-normal (term f)))
        "loopB-inner")))

; checks that ⇓ is a function, as claimed
; “The normalization relation ⇓ ... is indeed a function.”
(define (cc-normal e)
  (match (remove-duplicates (apply-reduction-relation ⇓ (weak-head-normal e)))
    [(list e’) e’]
    [es (error 'cc-normal "normalizing\n~a\nproduces ~a results:\n~a"
               (pretty-format e)
               (length es)
               (pretty-format es))]))

(define ccnf? (redex-match CCA ccnf))

; checks that ↦ is single-valued, as claimed
; “[The ↦ rules] are adjusted to avoid overlaps.”
; (Don't know whether the authors would consider 
; self-overlap to be overlap.)
(define (reduced e)
  (match (apply-reduction-relation ↦ e)
    [(list) (list)]
    [(list e’) (list e’)]
    [_ (error 'reduced "the following expression does not produce less than two results:\n~a"
              (pretty-format e))]))

(define β
  (reduction-relation
   CCA #:domain e
   (--> (fst (pair e_1 e_2)) e_1
        "fst")
   (--> (snd (pair e_1 e_2)) e_2
        "snd")
   (--> ((λ (x) e_1) e_2)
        (subst x e_2 e_1)
        "app")
   (--> (assoc e)
        (pair (fst (fst e)) (pair (snd (fst e)) (snd e)))
        "assoc")
   (--> (assoc-inv e)
        (pair (pair (fst e) (fst (snd e))) (snd (snd e)))
        "assoc-inv")
   (--> (swap e)
        (pair (snd e) (fst e))
        "swap")
   (--> (times e)
        (λ (g) (λ (z) (pair (e (fst z)) (g (snd z)))))
        (fresh g z)
        "times")
   (--> (compose e)
        (λ (g) (λ (x) (e (g x))))
        (fresh g x)
        "compose")
   (--> (id e)
        e
        "id")
   (--> (juggle e)
        (((compose ((compose assoc-inv) ((times id) swap))) assoc) e)
        "juggle")
   (--> (shuffle e)
        (((compose assoc-inv) ((times id) transpose)) e)
        "shuffle")
   (--> (shuffle-inv e)
        (((compose ((times id) transpose)) assoc) e)
        "shuffle-inv")
   (--> (second e)
        ((>>> ((>>> (arr swap)) (first e))) (arr swap))
        "second")
   #;(--> (loopB e)
        (λ (x) (loop ((>>> x) (second (second (init e))))))
        (fresh x)
        "loopB")))

(define (normal-form R e)
  (match (apply-reduction-relation R e)
    [(cons e’ _) (normal-form R e’)]
    [(list) e]))

(define weak-head-normal
  (let ()
    (define-extended-language CCA-E CCA
      (E hole 
         (E e)
         (fst E)
         (snd E)))
    (curry normal-form (context-closure β CCA-E E))))

(define-metafunction CCA 
  ;; 1. x_1 bound, so don't continue in λ body 
  [(subst x_1 any_1 (λ (x_1) any_2))
   (λ (x_1) any_2)]
  ;; 2. general purpose capture avoiding case 
  [(subst x_1 any_1 (λ (x_2) any_2))
   (λ (x_new) (subst x_1 any_1 (subst-var (x_2 x_new) any_2)))
   (where x_new ,(variable-not-in (term (x_1 any_1 any_2)) (term x_2)))]
  ;; 3. replace x_1 with e_1 
  [(subst x_1 any_1 x_1) any_1]
  ;; 4. x_1 and x_2 are different, so don't replace 
  [(subst x_1 any_1 x_2) x_2]
  ;; the last cases cover all other expressions 
  [(subst x_1 any_1 (any_2 ...))
   ((subst x_1 any_1 any_2) ...)]
  [(subst x_1 any_1 any_2) any_2])

(define-metafunction CCA
  [(subst-var (x_1 any_1) x_1) any_1]
  [(subst-var (x_1 any_1) (any_2 ...))
   ((subst-var (x_1 any_1) any_2) ...)]
  [(subst-var (x_1 any_1) any_2) any_2]
  [(subst-var (x_1 any_1) (x_2 any_2) ... any_3)
   (subst-var (x_1 any_1) (subst-var (x_2 any_2) ... any_3))])