#lang scheme

;; this module defines the semantics of the "target" machine. The paper calls this the
;; "distributed abstract state machine".

(require redex/reduction-semantics)

(provide (all-defined-out))

(define-language dasm ;; Distributed Abstract State Machine
  (s (s ...)
     (nu (p ...) s)       ;; name binding, a la pi calculus
     (active p a)         ;; communication waiting to happen
     (heart p)            ;; communication that has found a mate
     a                    ;; completed communication
     (free chan)          ;; an available channel
     (announced chan p p) ;; an announcement of a pairing of in & out
     (open sync)          ;; an open synchronizer (a.k.a. choose)
     (closed sync)        ;; a synchronizer that is trying a match
     (selected sync p)    ;; a communication with blessing from synchronizer
     (refused p)          ;; a communication not blessed by synchronizer
     (confirmed sync p)   ;; confirmed, ready for action (unneccesary?)
     (canceled sync)      ;; failed synchronizer, ready for reboot
     )
  
  (sync ((p a) ...))
  (a (in chan) (out chan)) ; same as source model
  (chan variable)
  (p variable))

(define (find-bad-s los)
  (ormap (lambda (s) (if (redex-match dasm s s) #f s)) los))

(define t->
  (reduction-relation 
   dasm
   
   ;;rule 4.i
   
   (==> (s_0 ... (confirmed ((p_0 a_0) ... (p a) (p_1 a_1) ...) p) s_1 ...)
        (s_0 ... a s_1 ...))
   
   ;; rule 4.ii
   
   (==> (s_0 ... (canceled ((p a) ...)) s_1 ...)
        (s_0 ... (nu (no-dups (p ...)) ((open ((p a) ...)) (active p a) ...)) s_1 ...))
   
   ;; rule 3.i
   
   (==> (s_0 ... (announced chan p_0 p_1) s_1 ... (selected sync_0 p_0) s_2 ... (selected sync_1 p_1) s_3 ...)
        (s_0 ... s_1 ... s_2 ... s_3 ... (confirmed sync_0 p_0) (confirmed sync_1 p_1)))
   
   (==> (s_0 ... (announced chan p_1 p_0) s_1 ... (selected sync_0 p_0) s_2 ... (selected sync_1 p_1) s_3 ...)
        (s_0 ... s_1 ... s_2 ... s_3 ... (confirmed sync_0 p_0) (confirmed sync_1 p_1)))
   
   ;; rule 3.ii
   
   (==> (s_0 ... (announced chan p_0 p_1) s_1 ... (selected sync_0 p_0) s_2 ... (refused p_1) s_3 ...)
        (s_0 ... s_1 ... s_2 ... s_3 ... (canceled sync_0)))
   
   ;; rule 3.iii
   
   (==> (s_0 ... (announced chan p_1 p_0) s_1 ... (selected sync_0 p_0) s_2 ... (refused p_1) s_3 ...)
        (s_0 ... s_1 ... s_2 ... s_3 ... (canceled sync_0)))
   
   ;; rule 3.iv
   
   (==> (s_0 ... (announced chan p_0 p_1) s_1 ... s_2 ... (refused p_0) (refused p_1) s_3 ...)
        (s_0 ... s_1 ... s_2 ... s_3 ...))
   
   (==> (s_0 ... (announced chan p_1 p_0) s_1 ... s_2 ... (refused p_0) (refused p_1) s_3 ...)
        (s_0 ... s_1 ... s_2 ... s_3 ...))
   
   ;; rule 2.i
   
   (==> (s_0 ... (heart p)                                        s_1 ... (open   ((p_0 a_0) ... (p a) (p_1 a_1) ...)) s_2 ...)
        (s_0 ... (selected ((p_0 a_0) ... (p a) (p_1 a_1) ...) p) s_1 ... (closed ((p_0 a_0) ... (p a) (p_1 a_1) ...)) s_2 ...))
   
   ;; rule 2.ii
   
   (==> (s_0 ... (heart p)   s_1 ... (closed ((p_0 a_0) ... (p a) (p_1 a_1) ...)) s_2 ...)
        (s_0 ... (refused p) s_1 ... (closed ((p_0 a_0) ... (p a) (p_1 a_1) ...)) s_2 ...))
   
   ;; rule 1
   
   (==> (s_0 ... (active p_1 (in chan)) s_1 ... (active p_2 (out chan)) s_2 ... (free chan)                          s_3 ...)
        (s_0 ... (heart p_1)            s_1 ... (heart p_2)             s_2 ... (announced chan p_1 p_2) (free chan) s_3 ...))
   
   
   with
   [(--> (nu (p ...) A) (tsort (nu (p ...) B)))
    (==> A B)]))


;; actives : extract active points from a sync
(define-metafunction dasm
  actives : sync -> (s ...)
  [(actives ((p a) ...))  ((active p a) ...)])

;; no-dups : take a list of p's, return the list without duplications
(define-metafunction dasm
  no-dups : (p ...) -> (p ...)
  [(no-dups ()) ()]
  [(no-dups (p p_0 ... p p_1 ...)) (no-dups (p p_0 ... p_1 ...))]
  [(no-dups (p p_0 ...)) (p p_1 ...)
                         (where (p_1 ...) (no-dups (p_0 ...)))])

;; tsort: arrange a term into the canonical form: one 'nu' at the top, everything in order.
(define-metafunction dasm
  tsort : s -> s
  [(tsort ())           (nu () ())]
  [(tsort (s_0))        (tsort s_0)]
  [(tsort (s_0 ...))    (nu-combine (tsort (s_1 ...)) (tsort (s_2 ...)))
                        (where ((s_1 ...) (s_2 ...)) ,(let* ([states (term (s_0 ...))]
                                                             [half-len (floor (/ (length states) 2))])
                                                        (list (take states half-len)
                                                              (drop states half-len))))]
  [(tsort (nu (p_0 ...) s)) (nu (no-dups (p_0 ... p_1 ...)) s_1)
                            (where (nu (p_1 ...) s_1) (tsort s))]
  ;; everything else:
  [(tsort s) (nu () (s))])


;; given two sorted states, combine them into one sorted state
(define-metafunction dasm
  nu-combine : s s -> s
  [(nu-combine s_0 s_1) (nu (p_0c ... p_1c ...) (interleave s_0c s_1c))
   (where (nu (p_0c ...) s_0c) (clean s_0 s_1))
   (where (nu (p_1c ...) s_1c) (clean s_1 s_0c)) ;; note asymmetry w/ prior line.
   ])


;; interleave : take two sequences and interleave them to join all of the corresponding pieces in the right order
(define-metafunction dasm
  interleave : (s ...) (s ...) -> (s ...)
  [(interleave ((active p_a0 (in chan_a0)) ... (active p_a11 (out chan_a11)) ... (heart p_a1) ... a_a2 ... (free chan_a3) ... (announced chan_a4 p_a4a p_a4b) ...
                                           (open sync_a5) ... (closed sync_a6) ... (selected sync_a7 p_a7) ... (refused p_a8) ...
                                           (confirmed sync_a9 p_a9) ... (canceled sync_a10) ...)
               ((active p_b0 (in chan_b0)) ... (active p_b11 (out chan_b11)) ... (heart p_b1) ... a_b2 ... (free chan_b3) ... (announced chan_b4 p_b4a p_b4b) ...
                                           (open sync_b5) ... (closed sync_b6) ... (selected sync_b7 p_b7) ... (refused p_b8) ...
                                           (confirmed sync_b9 p_b9) ... (canceled sync_b10) ...))
   ((active p_a0 (in chan_a0)) ... (active p_b0 (in chan_b0)) ... (active p_a11 (out chan_a11)) ... (active p_b11 (out chan_b11)) ...
                               (heart p_a1) ... (heart p_b1) ... a_b2 ... a_a2 ... (free chan_a3) ... (free chan_b3) ...
                               (announced chan_a4 p_a4a p_a4b) ... (announced chan_b4 p_b4a p_b4b) ... (open sync_a5) ... (open sync_b5) ... 
                               (closed sync_a6) ... (closed sync_b6) ... (selected sync_a7 p_a7) ... (selected sync_b7 p_b7) ... (refused p_a8) ... (refused p_b8) ...
                               (confirmed sync_a9 p_a9) ... (confirmed sync_b9 p_b9)  ... (canceled sync_a10) ... (canceled sync_b10) ...)])


;; let's try to damp down that constant factor by artificially preventing variable name collisions
(define count 0)
(define (freshmaker)
  (set! count (+ count 1))
  (string->symbol (format "blork-~e" count)))

;; clean: take a nu form, and ensure that none of its names capture free vars of another state
(define-metafunction dasm
  clean : s s -> s
  ;; loop to take care of all hits:
  [(clean (nu (p ...) s_0) s_1) (clean (nu ((rename1 p_bad p_new p) ...) (rename p_bad p_new s_0)) s_1)
                                (where (p_bad) (overlap (p ...) (free-vars s_1)))
                                (where p_new ,(variable-not-in (term ((free-vars s_1) (free-vars s_0))) (term p_bad)))]
  [(clean (nu (p ...) s_0) s_1) (nu (p ...) s_0)])

;; overlap: return a (variable) that occurs in both lists, or () otherwise
(define-metafunction dasm
  overlap : (p ...) (p ...) -> (p ...)
  #;[(overlap (p_0 ...) (p_1 ...)) ,(let ([ht (make-hasheq)])
                                    (for-each (lambda (p) (hash-set! ht p #t)) (term (p_0 ...)))
                                    (let loop ([p1s (term (p_1 ...))])
                                      (cond [(empty? p1s) '()] 
                                            ;; RIGHT HERE.  ALSO, LOOPING UP IN CLEAN MAY BE THE CULPRIT.
                                            ))
                                    (ormap (lambda (p) (hash-has-key? ht p)) (term (p_1 ...))))]
  [(overlap (p_bad p_0 ...) (p_1 ... p_bad p_2 ...)) (p_bad)]
  [(overlap (p_okay p_0 ...) (p_1 ...)) (overlap (p_0 ...) (p_1 ...))]
  [(overlap (p_0 ...) (p_1 ...)) ()])



;; rename
(define-metafunction dasm
  rename : p p s -> s
  [(rename p_f p_t (s ...)) ((rename p_f p_t s) ...)]
  [(rename p_f p_t (nu (p_0 ... p_f p_1 ...) s)) (nu (p_0 ... p_f p_1 ...) s)]
  [(rename p_f p_t (nu (p_0 ... p_t p_1 ...) s)) (rename p_f p_t (nu (p_0 ... p_new p_1 ...) (rename p_t p_new s)))
                                                 (side-condition (member (term p_f) (term (free-vars s))))]
  [(rename p_f p_t (nu (p ...) s)) (nu (p ...) (rename p_f p_t s))]
  [(rename p_f p_t (active p_f a))           (active p_t a)] 
  [(rename p_f p_t (heart p_f))              (heart p_t)]
  [(rename p_f p_t (announced chan p_0 p_1)) (announced chan (rename1 p_f p_t p_0) (rename1 p_f p_t p_1))]
  [(rename p_f p_t (open sync))              (open (rename/sync p_f p_t sync))]
  [(rename p_f p_t (closed sync))            (closed (rename/sync p_f p_t sync))]
  [(rename p_f p_t (selected sync p))        (selected (rename/sync p_f p_t sync)  (rename1 p_f p_t p))]
  [(rename p_f p_t (confirmed sync p))       (confirmed (rename/sync p_f p_t sync) (rename1 p_f p_t p))]
  [(rename p_f p_t (canceled sync))          (canceled (rename/sync p_f p_t sync))]
  ;; fall-through:
  [(rename p_f p_t s) s])

;; rename/sync : rename within a sync
(define-metafunction dasm
  rename/sync : p p sync -> sync
  [(rename/sync p_f p_t ((p a) ...)) (((rename1 p_f p_t p) a) ...)])

;; rename1 : rename a single name, if necessary
(define-metafunction dasm
  rename1 : p p p -> p
  [(rename1 p_f p_t p_f) p_t]
  [(rename1 p_f p_t p_t) (error "should never happen 2009-11-18 14:34")]
  [(rename1 p_f p_t p) p])


;; free-vars, with duplication
(define-metafunction dasm
  free-vars : s -> (p ...)
  [(free-vars (s ...)) (p ... ...)
                       (where ((p ...) ...) ((free-vars s) ...))]
  [(free-vars (nu (p ...) s)) (set-minus (free-vars s) (p ...))]
  [(free-vars (active p a))   (p)]
  [(free-vars (heart p))      (p)]
  [(free-vars (announced chan p_0 p_1)) (p_0 p_1)]
  [(free-vars (open sync))    (free-vars/sync sync)]
  [(free-vars (closed sync))  (free-vars/sync sync)]
  [(free-vars (selected sync p))        (p p_s ...)
                                        (where (p_s ...) (free-vars/sync sync))]
  [(free-vars (refused p))    (p)]
  [(free-vars (confirmed sync p)) (p p_s ...)
                                  (where (p_s ...) (free-vars/sync sync))]
  [(free-vars (canceled sync))    (free-vars/sync sync)]
  ;; fall-through
  [(free-vars s)                  ()])

(define-metafunction dasm
  free-vars/sync : sync -> (p ...)
  [(free-vars/sync ((p a) ...)) (p ...)])


;; remove *all* instances of the second set from the first
(define-metafunction dasm
  set-minus : (p ...) (p ...) -> (p ...)
  [(set-minus (p_1 ...) (p_2 ...)) ,(remove* (term (p_2 ...)) (term (p_1 ...)))])



;; denotes : following the definition in the paper, a machine state denotes the 
;; set of completed communication events.
(define-metafunction dasm 
  denotes/t : s -> (a ...)
  [(denotes/t (s ...)) (a ... ...)
                       (where ((a ...) ...) ((denotes/t s) ...))]
  [(denotes/t (nu (p ...) s)) (denotes/t s)]
  [(denotes/t a) (a)]
  ;; catch-all:
  [(denotes/t s) ()])